#####################################################################################################
# Python script for extracting elevation values from lidar data for a string of points              #
# Released under the terms of GPLv3 by acrallycentral.com                                           #
# Version 1.1 - 26/09/2024                                                                          #
#                                                                                                   #
# Changelog:                                                                                        #
# v1.1: remove accidentaly hardcoded EPSG code                                                      #
#                                                                                                   #
# To install dependencies: run following commands                                                   #
# > conda install conda-forge::pdal gpxpy                                                           #
# > pip install geopandas shapely numpy pandas pyproj matplotlib tqdm                               #
#####################################################################################################

import gpxpy
import geopandas as gpd
from shapely.geometry import Point, LineString, box
import numpy as np
import pandas as pd
import logging
from pyproj import CRS, Transformer
import pdal
import json
import os
import sys
import matplotlib.pyplot as plt
from concurrent.futures import ThreadPoolExecutor, as_completed
from tqdm import tqdm
from functools import lru_cache
import configparser
import glob
import multiprocessing

# Set up logging
logging.basicConfig(level=logging.INFO, format='%(asctime)s - %(levelname)s - %(message)s', datefmt='%H:%M:%S')

def resource_path(relative_path):
    """Get absolute path to resource, works for dev and for PyInstaller"""
    try:
        base_path = sys._MEIPASS
    except Exception:
        base_path = os.path.abspath(".")
    return os.path.join(base_path, relative_path)

def load_config():
    """Load configuration from INI file"""
    config = configparser.ConfigParser(inline_comment_prefixes='#')
    application_path = os.path.dirname(sys.executable) if getattr(sys, 'frozen', False) else os.path.dirname(os.path.abspath(__file__))
    config_path = resource_path('GPX-to-points-string-with-elevations.ini')
    logging.info(f"Looking for config file at: {config_path}")
    if not os.path.exists(config_path):
        raise FileNotFoundError(f"Configuration file not found: {config_path}")
    config.read(config_path)
    return config, application_path

def get_num_threads(config):
    """Determine the number of threads to use"""
    num_threads = config.getint('Processing', 'NUM_THREADS')
    fallback_threads = config.getint('Processing', 'FALLBACK_THREADS', fallback=8)
    
    if num_threads == 0:  # Auto-detect
        try:
            num_threads = multiprocessing.cpu_count()
            logging.info(f"Auto-detected {num_threads} CPU threads")
        except:
            num_threads = fallback_threads
            logging.warning(f"Failed to auto-detect CPU cores. Using fallback value: {num_threads}")
    else:
        logging.info(f"Using {num_threads} threads as specified in the configuration")
    
    return min(num_threads, fallback_threads)  # Ensure we don't exceed the fallback limit

def get_file_paths(config, application_path):
    """Get file paths from config"""
    input_file = os.path.abspath(os.path.join(application_path, config.get('Files', 'INPUT_FILE')))
    input_base = os.path.splitext(os.path.basename(input_file))[0]
    return {
        'input_file': input_file,
        'copc_file': os.path.abspath(os.path.join(application_path, config.get('Files', 'COPC_FILE'))),
        'copc_dir': os.path.abspath(os.path.join(application_path, config.get('Files', 'COPC_DIR'))),
        'output_gpkg': os.path.join(application_path, config.get('Output', 'OUTPUT_GPKG').format(input_base=input_base)),
        'output_csv': os.path.join(application_path, config.get('Output', 'OUTPUT_CSV').format(input_base=input_base)),
        'output_png': os.path.join(application_path, config.get('Output', 'OUTPUT_PNG').format(input_base=input_base))
    }

def check_files_and_folders(file_paths, multi_file):
    """Check for the presence of required files and folders"""
    if not os.path.exists(file_paths['input_file']):
        raise FileNotFoundError(f"Input GPX file not found: {file_paths['input_file']}")

    if multi_file:
        if not os.path.isdir(file_paths['copc_dir']):
            raise FileNotFoundError(f"COPC directory not found: {file_paths['copc_dir']}")
        copc_files = glob.glob(os.path.join(file_paths['copc_dir'], '*.copc.laz'))
        if not copc_files:
            raise FileNotFoundError(f"No COPC files found in directory: {file_paths['copc_dir']}")
    else:
        if not os.path.exists(file_paths['copc_file']):
            raise FileNotFoundError(f"COPC file not found: {file_paths['copc_file']}")

    logging.info("All required files and folders are present.")

def create_points_at_equal_distances(line, distance):
    """Create points at equal distances along a line"""
    points = []
    current_distance = 0
    line_length = line.length
    while current_distance <= line_length:
        points.append(line.interpolate(current_distance))
        current_distance += distance
    if points[-1].distance(Point(line.coords[-1])) > 1e-8:
        points.append(Point(line.coords[-1]))
    return points

def idw_interpolation(points, values, target_point, power=2):
    """Perform IDW interpolation"""
    distances = np.linalg.norm(points - target_point, axis=1)
    weights = 1 / (distances**power)
    return np.sum(values * weights) / np.sum(weights)

@lru_cache(maxsize=None)
def get_copc_info(copc_file, default_crs):
    """Retrieve COPC file information"""
    if not os.path.exists(copc_file) or not os.access(copc_file, os.R_OK):
        logging.error(f"COPC file not found or not readable: {copc_file}")
        return None, None

    try:
        pipeline = pdal.Pipeline(json.dumps([{"type": "readers.copc", "filename": copc_file, "count": 0}]))
        pipeline.execute()
        metadata = pipeline.metadata

        if isinstance(metadata, dict):
            metadata_json = metadata
        else:
            metadata_json = json.loads(metadata)

        copc_info = metadata_json.get('metadata', {}).get('readers.copc', {})
        
        if not copc_info:
            logging.error("Could not find COPC info in metadata")
            return None, None

        bounds = copc_info.get('bounds') or copc_info.get('minx')
        if not bounds:
            logging.error("Could not find bounds information in metadata")
            return None, None

        formatted_bounds = bounds if isinstance(bounds, dict) else {
            'minx': copc_info.get('minx'), 'miny': copc_info.get('miny'),
            'maxx': copc_info.get('maxx'), 'maxy': copc_info.get('maxy')
        }

        srs = copc_info.get('srs', {})
        if srs and isinstance(srs, dict) and 'wkt' in srs:
            try:
                crs = CRS.from_wkt(srs['wkt'])
                logging.info(f"EPSG code found in COPC file: {crs.to_epsg()}")
            except Exception as e:
                logging.warning(f"No EPSG code found in COPC file. Using default CRS: EPSG:{default_crs}")
                crs = CRS.from_epsg(default_crs)
        else:
            logging.warning(f"No valid SRS found in COPC file. Using default CRS: EPSG:{default_crs}")
            crs = CRS.from_epsg(default_crs)

        return crs, formatted_bounds
    except Exception as e:
        logging.error(f"Error retrieving information from COPC file: {str(e)}")
        logging.error(f"Full metadata: {metadata}")
        return None, None

def get_elevation(copc_file, x, y, copc_crs, search_radius, num_points, use_idw, classified):
    """Get elevation for a point"""
    def execute_pipeline(radius):
        min_x, min_y = x - radius, y - radius
        max_x, max_y = x + radius, y + radius
        json_pipeline = [
            {
                "type": "readers.copc",
                "filename": copc_file,
                "bounds": f"([{min_x}, {max_x}], [{min_y}, {max_y}])"
            }
        ]
        
        if classified:
            json_pipeline.append({
                "type": "filters.range",
                "limits": "Classification[2:2],Classification[11:11],Classification[17:17]"
            })
        
        pipeline = pdal.Pipeline(json.dumps(json_pipeline))
        return pipeline, pipeline.execute()

    pipeline, count = execute_pipeline(search_radius)
    
    if count == 0:
        for _ in range(5):
            search_radius *= 2
            pipeline, count = execute_pipeline(search_radius)
            if count > 0:
                break
        else:
            raise ValueError(f"No points found within expanded search area")

    point_data = pipeline.arrays[0]
    points = np.column_stack((point_data['X'], point_data['Y']))
    elevations = point_data['Z']
    
    distances = np.linalg.norm(points - np.array([x, y]), axis=1)
    nearest_indices = np.argpartition(distances, num_points)[:num_points]
    
    nearest_points = points[nearest_indices]
    nearest_elevations = elevations[nearest_indices]
    
    return idw_interpolation(nearest_points, nearest_elevations, np.array([x, y])) if use_idw else np.mean(nearest_elevations)

def create_height_profile(distances, elevations, output_png):
    """Create and save height profile plot"""
    plt.figure(figsize=(12, 6))
    plt.plot(distances, elevations)
    plt.title('Height Profile')
    plt.xlabel('Distance (m)')
    plt.ylabel('Elevation (m)')
    plt.grid(True)
    plt.tight_layout()
    plt.savefig(output_png, format='png', dpi=300)
    plt.close()
    logging.info(f"Height profile saved as {output_png}")

def get_copc_files(copc_file, copc_dir, multi_file):
    """Get list of COPC files based on configuration"""
    if multi_file:
        return glob.glob(os.path.join(copc_dir, '*.copc.laz'))
    else:
        return [copc_file]

def process_gpx_with_elevation(input_file, output_gpkg, output_csv, output_png, sample_distance, copc_files, default_crs, include_distance, search_radius, num_points, use_idw, classified, num_threads):
    """Main processing function"""
    with open(input_file, 'r') as gpx_file:
        gpx = gpxpy.parse(gpx_file)
    
    track_points = [Point(point.longitude, point.latitude) for track in gpx.tracks for segment in track.segments for point in segment.points]
    logging.info(f"Number of original track points: {len(track_points)}")
    
    if len(track_points) < 2:
        raise ValueError("The GPX file must contain at least two points")
    
    gdf = gpd.GeoDataFrame(geometry=track_points, crs="EPSG:4326")
    
    # Get CRS and bounds for all COPC files
    copc_info = []
    for copc_file in copc_files:
        crs, bounds = get_copc_info(copc_file, default_crs)
        if crs is not None and bounds is not None:
            copc_info.append((copc_file, crs, bounds))
    
    if not copc_info:
        logging.error("Failed to retrieve COPC file information. Exiting.")
        return

    # Use the CRS from the first COPC file
    copc_crs = copc_info[0][1]
    
    # Create a GeoDataFrame with all COPC bounding boxes
    copc_boxes = [box(bounds['minx'], bounds['miny'], bounds['maxx'], bounds['maxy']) for _, _, bounds in copc_info]
    copc_gdf = gpd.GeoDataFrame(geometry=copc_boxes, crs=copc_crs)
    
    # Reproject GPX points to COPC CRS
    gdf_reprojected = gdf.to_crs(copc_crs)
    
    # Check if all points are within any COPC bounding box
    points_within = gdf_reprojected.geometry.apply(lambda point: any(point.within(box) for box in copc_boxes))
    if not points_within.all():
        logging.warning("Some GPX track points are not contained within any COPC file's bounding box.")
        logging.warning("This may indicate some tiles are missing, or incorrect EPSG code.")
        logging.warning("Please check the EPSG code, lidar tiles and GPX in QGIS.")
        logging.warning("Stopping processing to prevent potential errors.")
        return
    
    # Use the COPC CRS instead of hardcoded UTM
    gdf_copc_crs = gdf.to_crs(copc_crs)
    line_copc_crs = LineString(gdf_copc_crs.geometry.tolist())
    logging.info(f"Original track length: {line_copc_crs.length:.2f} meters")
    
    equal_distance_points_copc_crs = create_points_at_equal_distances(line_copc_crs, sample_distance)
    logging.info(f"Number of equally spaced points: {len(equal_distance_points_copc_crs)}")
    
    gdf_points_copc_crs = gpd.GeoDataFrame(geometry=equal_distance_points_copc_crs, crs=copc_crs)
    gdf_points = gdf_points_copc_crs.to_crs("EPSG:4326")
    
    transformer = Transformer.from_crs(CRS.from_epsg(4326), copc_info[0][1], always_xy=True)
    
    def get_elevation_for_point(args):
        index, point = args
        x, y = transformer.transform(point.x, point.y)
        
        # Find the appropriate COPC file for this point
        for copc_file, crs, bounds in copc_info:
            if bounds['minx'] <= x <= bounds['maxx'] and bounds['miny'] <= y <= bounds['maxy']:
                elevation = get_elevation(copc_file, x, y, crs, search_radius, num_points, use_idw, classified)
                return index, elevation
        
        # If no COPC file contains this point, return None for elevation
        return index, None

    with ThreadPoolExecutor(max_workers=num_threads) as executor:
        future_to_index = {executor.submit(get_elevation_for_point, (i, point)): i for i, point in enumerate(gdf_points.geometry)}
        elevation_results = {}
        for future in tqdm(as_completed(future_to_index), total=len(future_to_index), desc="Fetching elevations"):
            index, elevation = future.result()
            elevation_results[index] = elevation

    elevations = [elevation_results[i] for i in range(len(gdf_points))]
    gdf_points['elevation'] = elevations
    gdf_points['cumulative_distance'] = gdf_points_copc_crs.distance(gdf_points_copc_crs.shift()).cumsum().fillna(0)
    
    gdf_points.to_file(output_gpkg, driver="GPKG", layer='track_points')
    logging.info(f"Processed GPX file with elevations saved as {output_gpkg}")
    
    csv_data = pd.DataFrame({
        'longitude': gdf_points.geometry.x,
        'latitude': gdf_points.geometry.y,
        'elevation': gdf_points.elevation.round(2),
    })

    if include_distance:
        csv_data['cumulative_distance'] = gdf_points.cumulative_distance.round(2)

    csv_data.to_csv(output_csv, index=False)
    logging.info(f"Processed GPX file with elevations{'and cumulative distances' if include_distance else ''} saved as {output_csv}")

    create_height_profile(gdf_points['cumulative_distance'], csv_data['elevation'], output_png)

if __name__ == "__main__":
    try:
        config, application_path = load_config()
        file_paths = get_file_paths(config, application_path)
        multi_file = config.getboolean('Processing', 'MULTI_FILE')

        # Check for the presence of required files and folders at the start
        check_files_and_folders(file_paths, multi_file)

        copc_files = get_copc_files(file_paths['copc_file'], file_paths['copc_dir'], multi_file)
        num_threads = get_num_threads(config)

        process_gpx_with_elevation(
            file_paths['input_file'], file_paths['output_gpkg'], file_paths['output_csv'], file_paths['output_png'],
            config.getfloat('Processing', 'POINTS_DISTANCE'),
            copc_files,
            config.getint('Processing', 'DEFAULT_CRS'),
            config.getboolean('Processing', 'INCLUDE_DISTANCE'),
            config.getfloat('Processing', 'SEARCH_RADIUS'),
            config.getint('Processing', 'NUM_POINTS'),
            config.getboolean('Processing', 'USE_IDW'),
            config.getboolean('Processing', 'CLASSIFIED'),
            num_threads
        )
    except FileNotFoundError as e:
        logging.error(f"File not found error: {str(e)}")
    except Exception as e:
        logging.error(f"An error occurred: {str(e)}", exc_info=True)
    finally:
        if config.getboolean('Processing', 'WAIT_FOR_KEY', fallback=True):
            print("\nPress any key to exit...")
            input()
        else:
            print("\nProcessing complete. Exiting...")